from argparse import ArgumentParser
from fnmatch import fnmatch
from json import dumps
from functools import reduce
from pathlib import Path
from sys import exit
from typing import Optional

from pyiddidf.exceptions import ProcessingException
from pyiddidf.idd_objects import IDDObject
from pyiddidf.idd_processor import IDDProcessor


class Actions:
    IDDCheck = 'idd_check'
    FindIDDObjectsMatching = 'find_idd_objects_matching'
    SummarizeIDDObject = 'summarize_idd_object'


class ExitCodes:
    OK = 0
    ProcessingError = 1
    BadArguments = 2


# Eventually this could become a more feature rich CLI
def main_cli() -> int:
    parser = ArgumentParser(
        prog="energyplus_idd_idf",
        description="EnergyPlus IDD/IDF Utility Command Line",
        epilog="This CLI is in infancy and will probably have features added over time"
    )
    parser.add_argument('filename', help="Path to IDD/IDF file to be operated upon")  # positional argument
    parser.add_argument(
        '--idd_check', action='store_const', const=Actions.IDDCheck,
        help="Process the given IDD file and report statistics and issues"
    )
    parser.add_argument(
        '--idd_obj_matches', type=str, help="Find IDD objects that match the given basic pattern"
    )
    parser.add_argument(
        '--summarize_idd_object', type=str, help="Print a summary of a single IDD object by name"
    )
    args = parser.parse_args()
    all_options = [
        args.idd_check, args.idd_obj_matches, args.summarize_idd_object
    ]
    if all([x is None for x in all_options]):
        print(dumps({'message': "Nothing to do...use command line switches to perform operations"}, indent=2))
        return ExitCodes.OK
    p = Path(args.filename)
    if not p.exists():
        print(dumps({'message': "Supplied file does not appear to exist, check paths and retry!"}, indent=2))
        return ExitCodes.BadArguments
    # for now assume it's always the IDD so we don't have to repeat this code
    processor = IDDProcessor()
    try:
        processor.process_file_given_file_path(str(p))
    except ProcessingException:
        print("Issues occurred during processing")
        return ExitCodes.ProcessingError
    if args.idd_check:
        num_groups = len(processor.idd.groups)
        num_objects = reduce(
            lambda x, y: x + y,
            [len(g.objects) for g in processor.idd.groups],
            0
        )
        print(dumps({
            'message': 'Everything looks OK',
            'content': {
                'idd_version': processor.idd.version_string,
                'idd_build_id': processor.idd.build_string,
                'num_groups': num_groups,
                'num_objects': num_objects
            }
        }, indent=2))
    elif args.idd_obj_matches:
        pattern = args.idd_obj_matches
        matching_objects = []
        for g in processor.idd.groups:
            for o in g.objects:
                obj_name = o.name
                if fnmatch(obj_name, pattern):
                    matching_objects.append(o)
        print(dumps({
            'message': 'Everything looks OK',
            'content': {
                'pattern': pattern,
                'matching_objects': [
                    o.name for o in matching_objects
                ]
            }
        }, indent=2))
    elif args.summarize_idd_object:
        object_name = args.summarize_idd_object.upper()
        matching_object: Optional[IDDObject] = None
        for g in processor.idd.groups:
            for o in g.objects:
                obj_name = o.name.upper()
                if fnmatch(obj_name, object_name):
                    matching_object = o
        if matching_object is None:
            print(dumps({'message': f"Could not find matching object by name {object_name}"}, indent=2))
            return ExitCodes.BadArguments
        print(dumps({
            'message': 'Everything looks OK',
            'content': {
                'searched_object_name': object_name,
                'field': [
                    f"{f.field_an_index} : {f.field_name}" for f in matching_object.fields
                ]
            }
        }, indent=2))
    return ExitCodes.OK


if __name__ == "__main__":  # pragma: no cover
    exit(main_cli())



class UnimplementedMethodException(Exception):
    """
    This exception occurs when a call is made to a function that should be implemented in a derived class
    but is not, so the base class function is called.  This is a developer issue.

    :param str class_name: The name of the base class where the virtual function is defined
    :param str method_name: The method name which should be overridden in the derived class
    """

    def __init__(self, class_name, method_name):
        self.class_name = class_name
        self.method_name = method_name

    def __str__(self):
        return "{} derived classes should override {}() method".format(self.class_name, self.method_name)


class FileAccessException(Exception):
    """
    This exception occurs when the transition tool encounters a problem accessing a prescribed input or output file.

    :param str file_path: The file path which is causing the issue
    :param str problem_type: The type of problem occurring, from the constants defined in this class
    :param str file_nickname: The nickname of the file, from the constants defined in this class
    :param str message: An optional additional message to write out
    """

    CANNOT_FIND_FILE = "cannot find file"
    CANNOT_READ_FILE = "cannot read file"
    CANNOT_WRITE_TO_FILE = "cannot write to file"
    FILE_EXISTS_MUST_DELETE = "file exists, must delete"
    TRIED_BUT_CANNOT_DELETE_FILE = "tried to delete file, but couldn't"

    ORIGINAL_INPUT_FILE = "original input file"
    UPDATED_INPUT_FILE = "updated input file"
    ORIGINAL_DICT_FILE = "original dictionary file"
    UPDATED_DICT_FILE = "updated dictionary file"

    def __init__(self, file_path, problem_type, file_nickname, message=None):
        self.file_path = file_path
        self.problem_type = problem_type
        self.file_nickname = file_nickname
        self.message = message

    def __str__(self):
        s = "File access problem occurred:\n  Trying to operate on *{}* at \"{}\"\n  Problem: {}".format(
            self.file_nickname, self.file_path, self.problem_type)
        if self.message:
            s += "  Message: " + self.message
        return s


class FileTypeException(Exception):
    """
    This exception occurs when the prescribed file types do not match the expected conditions.
    """

    ORIGINAL_INPUT_FILE = "original input file"
    UPDATED_INPUT_FILE = "updated input file"
    ORIGINAL_DICT_FILE = "original dictionary file"
    UPDATED_DICT_FILE = "updated dictionary file"

    def __init__(self, file_path, file_nickname, message):
        self.file_path = file_path
        self.file_nickname = file_nickname
        self.message = message

    def __str__(self):
        return "File type problem occurred:\n  Trying to operate on *{}* at \"{}\"\n  Problem: {}".format(
            self.file_nickname, self.file_path, self.message
        )


class ManagerProcessingException(Exception):
    """
    This exception occurs when the transition tool encounters an unexpected issue when doing the transition.
    """

    def __init__(self, msg, issues=None):
        self.message = msg
        self.issues = issues

    def __str__(self):
        msg = ""
        if self.issues:
            for i in self.issues:
                msg += str(i) + "\n"
        msg += self.message
        return msg


class ProcessingException(Exception):
    """
    This exception occurs when an unexpected error occurs during the processing of an input file.
    """

    def __init__(self, message, line_index=None, object_name="", field_name=""):
        super(ProcessingException, self).__init__(message)
        self.message = message
        self.line_index = line_index
        self.object_name = object_name

    def __str__(self):
        if self.line_index:
            return f"Processing Exception on line number {self.line_index}; message: {self.message} (tentative object" \
                   f" name: \"{self.object_name}\" "
        else:
            return f"Processing Exception; message: {self.message}"



from io import StringIO
import logging
import os

from pyiddidf import exceptions
from pyiddidf.idf_objects import IDFObject, IDFStructure

module_logger = logging.getLogger("eptransition.idd.processor")


class IDFProcessor:
    """
    The core IDF Processor class.  Given an IDF via stream or path, this class has workers to robustly process the IDF
    into a rich IDFStructure instance.

    The constructor takes no arguments but sets up instance variables. Relevant "public" members are listed here:

    :ivar IDFStructure idf: The resulting IDFStructure instance after processing the IDF file/stream
    :ivar str file_path: A file path for this IDF, although it may be just a simple descriptor
    """

    def __init__(self):
        self.idf = None
        self.file_path = None
        self.input_file_stream = None

    def process_file_given_file_path(self, file_path):
        """
        This worker allows processing of an IDF file at a specific path on disk.

        :param file_path: The path to an IDF file on disk.
        :return: An IDFStructure instance created from processing the IDF file
        :raises ProcessingException: if the specified file does not exist
        """
        if not os.path.exists(file_path):
            raise exceptions.ProcessingException("Input file not found=\"" + file_path + "\"")
        self.input_file_stream = open(file_path, "r")
        self.file_path = file_path
        return self.process_file()

    def process_file_via_stream(self, idf_file_stream):
        """
        This worker allows processing of an IDF snippet via stream.  Most useful for unit testing, but possibly for
        other situations.

        :param file-like-object idf_file_stream: An IDF snippet that responds to typical file-like commands such as
                                                 read().  A common object would be the StringIO object.
        :return: An IDFStructure instance created from processing the IDF snippet
        """
        self.input_file_stream = idf_file_stream
        self.file_path = "/streamed/idf"
        return self.process_file()

    def process_file_via_string(self, idf_string):
        """
        This worker allows processing of an IDF snippet string.  Most useful for unit testing, but possibly for
        other situations.

        :param str idf_string: An IDF snippet string
        :return: An IDFStructure instance created from processing the IDF string
        """
        self.input_file_stream = StringIO(idf_string)
        self.file_path = "/string/idf/snippet"
        return self.process_file()

    def process_file(self):
        """
        Internal worker function that reads the IDF stream, whether it was constructed from a file path, stream or
        string.  This processor then processes the file line by line looking for IDF objects and comment blocks, and
        parsing them into a meaningful structure

        :return: An IDF structure describing the IDF contents
        :raises ProcessingException: for any issues encountered during the processing of the idf
        """
        self.idf = IDFStructure(self.file_path)
        # phase 0: read in lines of file
        lines = self.input_file_stream.readlines()

        class Blob:
            COMMENT = 1
            OBJECT = 2

            def __init__(self, blob_type, blob_lines=None):
                self.blob_type = blob_type
                if blob_lines is None:
                    blob_lines = []
                self.lines = blob_lines

        # should there maybe be 3 "state as of last line" options?
        # 1. reading comment block
        # 2. reading IDF block

        # so let's try keeping the idf in blobs of either comment data or object data
        current_blob = None
        initial_blobs = []
        for line in lines:
            line_text = line.strip()
            if len(line_text) == 0:
                continue
            elif line_text.startswith("!"):
                if current_blob is None:
                    current_blob = Blob(Blob.COMMENT)
                    current_blob.lines.append(line_text)
                elif current_blob.blob_type == Blob.COMMENT:
                    # just add to the lines and carry on
                    current_blob.lines.append(line_text)
                elif current_blob.blob_type == Blob.OBJECT:
                    # ignore it, we are still trying to read the object..
                    continue
            else:
                if current_blob is None:
                    # then this blob is fresh and is the start of a new object, but it could also be the end (one-liner)
                    current_blob = Blob(Blob.OBJECT)
                    actual_line = line_text
                    if "!" in line_text:
                        actual_line = line_text[:line_text.find("!")]
                    if ";" in actual_line:
                        # we end this object blob
                        current_blob.lines.append(line_text)
                        initial_blobs.append(current_blob)
                        current_blob = None
                    else:
                        current_blob.lines.append(line_text)
                elif current_blob.blob_type == Blob.OBJECT:
                    # then we should append this line to the current blob, but we also need to check if it is the end
                    current_blob.lines.append(line_text)
                    actual_line = line_text
                    if "!" in line_text:
                        actual_line = line_text[:line_text.find("!")]
                    if ";" in actual_line:
                        # we end this object blob
                        initial_blobs.append(current_blob)
                        current_blob = None
                elif current_blob.blob_type == Blob.COMMENT:
                    # then we need to package up the previous comment blob, and create this new one, but again..1-liner?
                    initial_blobs.append(current_blob)
                    current_blob = Blob(Blob.OBJECT)
                    current_blob.lines.append(line_text)
                    actual_line = line_text
                    if "!" in line_text:
                        actual_line = line_text[:line_text.find("!")]
                    if ";" in actual_line:
                        # we end this object blob
                        initial_blobs.append(current_blob)
                        current_blob = None
        if current_blob is not None:
            initial_blobs.append(current_blob)

        # next let's go blob by blob and clean up any trailing comments and such
        idf_objects = []
        for initial_blob in initial_blobs:
            if initial_blob.blob_type == Blob.COMMENT:
                idf_objects.append(IDFObject(initial_blob.lines, True))
            else:
                out_lines = []
                for line in initial_blob.lines:
                    line_text = line.strip()
                    this_line = ""
                    if len(line_text) > 0:
                        exclamation = line_text.find("!")
                        if exclamation == -1:
                            this_line = line_text
                        elif exclamation > 0:
                            this_line = line_text[:exclamation]
                        if not this_line == "":
                            out_lines.append(this_line.strip())
                # check these object lines for malformed idf syntax
                for li in out_lines:
                    if not (li.endswith(",") or li.endswith(";")):
                        raise exceptions.ProcessingException(
                            "IDF line doesn't end with comma/semicolon\nline:\"" + li + "\"")
                # intermediate: join entire array and re-split by semicolon
                idf_data_joined = "".join(out_lines)
                idf_object_strings = idf_data_joined.split(";")
                # phase 3: inspect each object and its fields
                for obj in idf_object_strings:
                    tokens = obj.split(",")
                    nice_object = [t.strip() for t in tokens]
                    if len(nice_object) == 1:
                        if nice_object[0] == "":
                            continue
                    idf_objects.append(IDFObject(nice_object))

        self.idf.objects = idf_objects

        try:
            self.idf.version_string = self.idf.get_idf_objects_by_type("Version")[0].fields[0]
            parse_version = True
        except IndexError:
            self.idf.version_string = 'UNKNOWN VERSION'
            parse_version = False
        if parse_version:
            try:
                version_tokens = self.idf.version_string.split(".")
                tmp_string = "{}.{}".format(version_tokens[0], version_tokens[1])
                self.idf.version_float = float(tmp_string)
            except ValueError:
                raise exceptions.ProcessingException(
                    "Found IDF version, but could not coerce into floating point representation")
        else:
            self.idf.version_float = 0.0
        return self.idf



from io import StringIO
import logging
import os
from typing import Optional

from pyiddidf import exceptions
from pyiddidf.idd_objects import IDDField, IDDObject, IDDStructure, IDDGroup

module_logger = logging.getLogger("eptransition.idd.processor")


class CurrentReadType:
    """
    Internal class containing constants for the different states of the actual IDD Processor engine
    """
    EncounteredComment_ReadToCR = 0
    ReadAnything = 1
    ReadingGroupDeclaration = 2
    ReadingObjectName = 3
    LookingForObjectMetaDataOrNextField = 4
    ReadingObjectMetaData = 5
    ReadingObjectMetaDataContents = 6
    ReadingFieldANValue = 7
    ReadingFieldMetaData = 8
    ReadingFieldMetaDataOrNextANValueOrNextObject = 9
    LookingForFieldMetaDataOrNextObject = 10
    LookingForFieldMetaDataOrNextField = 11


# keep a global dictionary of read IDD structures, could eventually move into the class, but right now we instantiate
# the class over and over so that wouldn't work
IDD_CACHE = {}


class IDDProcessor:
    """
    The core IDD Processor class.  Given an IDD via stream or path, this class has workers to robustly process the IDD
    into a rich IDDStructure instance.

    The constructor takes no arguments but sets up instance variables. Relevant "public" members are listed here:

    :ivar IDDStructure idd: The resulting IDDStructure instance after processing the IDD file/stream
    :ivar str file_path: A file path for this IDD, although it may be just a simple descriptor
    """

    def __init__(self):
        self.idd: Optional[IDDStructure] = None
        self.idd_file_stream = None
        self.file_path = None
        self.group_flag_string = "\\group"
        self.obj_flags = ["\\memo", "\\unique-object", "\\required-object", "\\min-fields",
                          "\\obsolete", "\\extensible", "\\format"]
        self.field_flags = ["\\field", "\\note", "\\required-field", "\\begin-extensible", "\\unitsBasedOnField",
                            "\\units", "\\ip-units", "\\scheduleunits", "\\minimum", "\\maximum", "\\default",
                            "\\deprecated", "\\autosizable", "\\autocalculatable", "\\type", "\\retaincase",
                            "\\key", "\\object-list", "\\reference-class-name", "\\reference", "\\external-list"]

    def process_file_given_file_path(self, file_path):
        """
        This worker allows processing of an IDD file at a specific path on disk.

        :param file_path: The path to an IDD file on disk.
        :return: An IDDStructure instance created from processing the IDD file
        :raises ProcessingException: if the specified file does not exist
        """
        if not os.path.exists(file_path):
            raise exceptions.ProcessingException("Input IDD file not found=\"" + file_path + "\"")  # pragma: no cover
        self.idd_file_stream = open(file_path, "rb")
        self.file_path = file_path
        return self.process_file()

    def process_file_via_stream(self, idd_file_stream):
        """
        This worker allows processing of an IDD snippet via stream.  Most useful for unit testing, but possibly for
        other situations.

        :param file-like-object idd_file_stream: An IDD snippet that responds to typical file-like commands such as
                                                 read().  A common object would be the StringIO object.
        :return: An IDDStructure instance created from processing the IDD snippet
        """
        self.idd_file_stream = idd_file_stream
        self.file_path = "/streamed/idd"
        return self.process_file()

    def process_file_via_string(self, idd_string):
        """
        This worker allows processing of an IDD snippet string.  Most useful for unit testing, but possibly for
        other situations.

        :param str idd_string: An IDD snippet string
        :return: An IDDStructure instance created from processing the IDD string
        """
        self.idd_file_stream = StringIO(idd_string)
        self.file_path = "/string/idd/snippet"
        return self.process_file()

    def peek_one_char(self) -> str:
        """
        Internal worker function that reads a single character from the internal IDD stream but resets the stream to
        the former position

        :return: A single character, the one immediately following the cursor, or None if it can't peek ahead.
        """
        pos = self.idd_file_stream.tell()
        c = self.idd_file_stream.read(1)
        if isinstance(c, bytes):
            try:
                c = c.decode('utf-8', errors='strict')
            except UnicodeDecodeError:
                c = '‍'  # this should really only occur if the idd is weirdly, I struggled creating a test for it
        elif isinstance(c, str):
            c = c
        # shouldn't be able to hit any other types... else: something
        if c == "":
            c = None
        self.idd_file_stream.seek(pos)
        return c

    def read_one_char(self) -> str:
        """
        Internal worker function that reads a single character from the internal IDD stream, advancing the cursor.

        :return: A single character, the one immediately following the cursor, or None if it can't read.
        """
        c = self.idd_file_stream.read(1)
        if isinstance(c, bytes):
            try:
                c = c.decode('utf-8', errors='strict')
            except UnicodeDecodeError:
                c = '‍'  # this should really only occur if the idd is weirdly, I struggled creating a test for it
        elif isinstance(c, str):
            c = c
        # shouldn't be able to hit any other types... else: something
        if c == "":
            c = None
        return c

    def process_file(self):
        """
        Internal worker function that reads the IDD stream, whether it was constructed from a file path, stream or
        string.  This state machine worker moves character by character reading tokens and processing them into
        a meaningful IDD structure.

        :return: An IDD structure describing the IDD contents
        :raises ProcessingException: for any erroneous conditions encountered during processing
        """
        # flags and miscellaneous variables
        line_index = 1  # 1-based counter for the current line of the file
        last_field_for_object = False  # this will be the last field if a semicolon is encountered
        magic_cache_key = None

        # variables used as we are building the input structure
        self.idd = IDDStructure(self.file_path)  # empty overall IDD structure
        cur_group = None  # temporary placeholder for an IDD group
        cur_object = None  # temporary placeholder for an IDD object
        cur_field = None  # temporary placeholder for an IDD field
        cur_obj_meta_data_type = None  # temporary placeholder for the type of object metadata encountered

        # variables related to building and processing tokens
        token_builder = ""

        # state machine variables
        read_status = CurrentReadType.ReadAnything  # current state machine reading status
        revert_status_after_comment = None  # reading status before the comment, shift back to this after comment's done

        # loop continuously, the loop will exit when it is done
        while True:

            # update the next character
            just_read_char = self.read_one_char()
            if not just_read_char:
                break

            # update the peeked character
            peeked_char = self.peek_one_char()
            if not peeked_char:
                peeked_char = "\n"  # to simulate that the line ended

            # if we are on Windows, we may end up with "\r", so move the read and peeked characters forward once
            if peeked_char == "\r":  # pragma no cover -- we don't unit test on Windows, so this won't be caught
                # in this case, we currently have like peeked_char='\r' and just_read_char='A'
                # we need to keep the just_read_char as it is, but move peeked char forward
                currently_just_read_char = just_read_char
                self.read_one_char()  # go ahead and digest the \r
                just_read_char = currently_just_read_char
                peeked_char = self.peek_one_char()
                if not peeked_char:
                    peeked_char = "\n"

            # jump if we are at an EOL
            if just_read_char == "\n":
                # increment the counter
                line_index += 1

            # if we aren't already processing a comment, and we have a comment:
            #  don't append to the token builder, just set read status
            if read_status != CurrentReadType.EncounteredComment_ReadToCR:
                if just_read_char == "!":
                    if read_status != CurrentReadType.ReadingFieldMetaData:
                        read_status = CurrentReadType.EncounteredComment_ReadToCR
                else:
                    token_builder += just_read_char

            # clear a preceding line feed character from the token
            if just_read_char == "\n" and len(token_builder) == 1:
                token_builder = ""

            if read_status == CurrentReadType.ReadAnything:

                # this is the most general case where we are wandering through the IDD looking for whatever
                # the possibilities are: comments, group declaration, or object definition
                if peeked_char == "\\":  # starting a group name
                    read_status = CurrentReadType.ReadingGroupDeclaration
                elif peeked_char in [" ", "\n", "\t"]:  # don't do anything
                    pass
                elif peeked_char == "!":
                    revert_status_after_comment = read_status
                    read_status = CurrentReadType.EncounteredComment_ReadToCR
                else:  # should be alphanumeric, just start reading object name
                    read_status = CurrentReadType.ReadingObjectName

            elif read_status == CurrentReadType.ReadingGroupDeclaration:

                # for the group declarations, we will just check to see if the
                # line has ended since it should be on a single line
                # if it hasn't then just keep on as is, if it has, parse the group name out of it
                if peeked_char == "\n":
                    # first update the previous group
                    if cur_group is not None:
                        self.idd.groups.append(cur_group)
                    group_declaration = token_builder.strip()
                    group_flag_index = group_declaration.find(self.group_flag_string)
                    if group_flag_index == -1:  # pragma: no cover
                        # add error to error report
                        raise exceptions.ProcessingException(
                            "Group keyword not found where expected",
                            line_index=line_index)
                    else:
                        group_declaration = group_declaration[len(self.group_flag_string):]
                    cur_group = IDDGroup(group_declaration.strip())
                    token_builder = ""
                    read_status = CurrentReadType.ReadAnything  # to start looking for groups/objects/comments/whatever

            elif read_status == CurrentReadType.ReadingObjectName:

                # the object names could have several aspects
                # they could be a single line object, such as: "Lead Input;"
                # they could be the title of a multi field object, such as: "Version,"
                # and they could of course have comments at the end
                # for now I will assume that the single line objects can't have metadata
                # so read until either a comma or semicolon, also trap for errors if we reach the end of line or comment
                if peeked_char == ",":
                    object_title = token_builder.strip()
                    cur_object = IDDObject(object_title)
                    token_builder = ""
                    self.read_one_char()  # to clear the comma
                    read_status = CurrentReadType.LookingForObjectMetaDataOrNextField
                elif peeked_char == ";":
                    # since this whole object is a single line, we can just add it directly to the current group
                    object_title = token_builder
                    # this is added to single-line objects because CurGroup isn't instantiated yet, should fix
                    self.idd.single_line_objects.append(object_title.strip())
                    token_builder = ""  # to clear the builder
                    self.read_one_char()  # to clear the semicolon
                    read_status = CurrentReadType.ReadAnything
                elif peeked_char in ["\n", "!"]:  # pragma: no cover
                    raise exceptions.ProcessingException(
                        "An object name was not properly terminated by a comma or semicolon",
                        line_index=line_index)

            elif read_status == CurrentReadType.LookingForObjectMetaDataOrNextField:

                token_builder = ""
                if peeked_char == "\\":
                    read_status = CurrentReadType.ReadingObjectMetaData
                elif peeked_char in ["A", "N"]:
                    read_status = CurrentReadType.ReadingFieldANValue
                elif peeked_char == "!":
                    revert_status_after_comment = read_status
                    read_status = CurrentReadType.EncounteredComment_ReadToCR
                elif peeked_char == " ":
                    # just let it keep reading
                    pass
                elif peeked_char == "\n":
                    # just let it keep reading
                    pass

            elif read_status == CurrentReadType.ReadingObjectMetaData:

                if peeked_char in [" ", ":", "\n"]:
                    if token_builder in self.obj_flags:
                        cur_obj_meta_data_type = token_builder
                        token_builder = ""
                        if cur_obj_meta_data_type in ["\\required-object", "\\unique-object"]:
                            # these do not carry further data, stop reading now
                            if cur_obj_meta_data_type not in cur_object.meta_data:
                                string_list = [None]
                                cur_object.meta_data[cur_obj_meta_data_type] = string_list
                            else:  # pragma: no cover   -- strings already exist, this is not valid...
                                raise exceptions.ProcessingException(
                                    "Erroneous object meta data - repeated \"" + token_builder + "\"",
                                    line_index=line_index,
                                    object_name=cur_object.name)
                            cur_obj_meta_data_type = None
                            read_status = CurrentReadType.LookingForObjectMetaDataOrNextField
                        else:
                            # these will have the following data, just set the flag
                            read_status = CurrentReadType.ReadingObjectMetaDataContents
                    else:  # pragma: no cover
                        # token_builder = ""
                        raise exceptions.ProcessingException(
                            "Erroneous object meta data tag found",
                            line_index=line_index,
                            object_name=cur_object.name)
                else:
                    # just keep reading
                    pass

            elif read_status == CurrentReadType.ReadingObjectMetaDataContents:

                if peeked_char == "\n":
                    data = token_builder.strip()
                    # quick validation of some meta data
                    if cur_obj_meta_data_type == "\\min-fields":
                        try:
                            float(data)
                        except ValueError:
                            raise exceptions.ProcessingException(
                                "Erroneous meta data for min-fields, non-numeric number of fields? Weird...",
                                line_index=line_index,
                                object_name=cur_object.name
                            )
                    if cur_obj_meta_data_type not in cur_object.meta_data:
                        string_list = [data]
                        cur_object.meta_data[cur_obj_meta_data_type] = string_list
                    else:
                        string_list = cur_object.meta_data[cur_obj_meta_data_type]
                        string_list.append(data)
                        cur_object.meta_data[cur_obj_meta_data_type] = string_list
                    token_builder = ""
                    cur_obj_meta_data_type = None
                    read_status = CurrentReadType.LookingForObjectMetaDataOrNextField

            elif read_status == CurrentReadType.ReadingFieldANValue:

                if peeked_char in [",", ";"]:
                    cur_field = IDDField(token_builder.strip())
                    token_builder = ""
                    if peeked_char == ",":
                        last_field_for_object = False
                    elif peeked_char == ";":
                        last_field_for_object = True
                    read_status = CurrentReadType.ReadingFieldMetaDataOrNextANValueOrNextObject
                elif peeked_char == "\n":  # pragma: no cover
                    raise exceptions.ProcessingException(
                        "Blank or erroneous ""AN"" field index value",
                        line_index=line_index,
                        object_name=cur_object.name)

            elif read_status == CurrentReadType.ReadingFieldMetaDataOrNextANValueOrNextObject:

                if peeked_char == "\\":
                    token_builder = ""
                    read_status = CurrentReadType.ReadingFieldMetaData
                elif peeked_char in ["A", "N"]:
                    token_builder = ""
                    # this is hit when we have an "AN" value right after a previous AN value, so no meta-data is added
                    if cur_field.field_name is None:
                        cur_field.field_name = ""
                    cur_object.fields.append(cur_field)
                    read_status = CurrentReadType.ReadingFieldANValue
                # If we hit a newline while searching here, we are moving onto the next object
                elif peeked_char == '\n':
                    token_builder = ""
                    # this is hit when we end an object when an "AN" declaration and nothing after it
                    if cur_field.field_name is None:
                        cur_field.field_name = ""
                    cur_object.fields.append(cur_field)
                    cur_group.objects.append(cur_object)
                    read_status = CurrentReadType.ReadAnything

            elif read_status == CurrentReadType.ReadingFieldMetaData:

                if peeked_char == "\n":

                    # for this one, let's read all the way to the end of the line, then parse data
                    flag_found = next((x for x in self.field_flags if x in token_builder), None)
                    if flag_found:
                        data = token_builder[len(flag_found):]
                        # data needs to start with a space, otherwise things like: \fieldd My Field would be valid
                        if len(data) > 0:
                            if data[0] not in [" ", ">", "<"]:
                                raise exceptions.ProcessingException(
                                    "Invalid meta data, expected a space after the meta data specifier before the data",
                                    line_index=line_index,
                                    object_name=cur_object.name,
                                    field_name=cur_field.field_name
                                )
                        data = data.strip()
                        if flag_found == "\\field":
                            cur_field.field_name = data
                        else:
                            if flag_found not in cur_field.meta_data:
                                string_list = [data]
                                cur_field.meta_data[flag_found] = string_list
                            else:
                                string_list = cur_field.meta_data[flag_found]
                                string_list.append(data)
                                cur_field.meta_data[flag_found] = string_list
                    else:  # pragma: no cover
                        raise exceptions.ProcessingException(
                            "Erroneous field meta data entry found",
                            line_index=line_index,
                            object_name=cur_object.name,
                            field_name=cur_field.field_name)
                    token_builder = ""
                    if last_field_for_object:
                        read_status = CurrentReadType.LookingForFieldMetaDataOrNextObject
                    else:
                        read_status = CurrentReadType.LookingForFieldMetaDataOrNextField

                else:
                    # just keep reading
                    pass

            elif read_status == CurrentReadType.LookingForFieldMetaDataOrNextField:

                if peeked_char in ["A", "N"]:
                    token_builder = ""
                    cur_object.fields.append(cur_field)
                    read_status = CurrentReadType.ReadingFieldANValue
                elif peeked_char == "\\":
                    token_builder = ""
                    read_status = CurrentReadType.ReadingFieldMetaData
                elif peeked_char == "!":
                    revert_status_after_comment = read_status
                    read_status = CurrentReadType.EncounteredComment_ReadToCR
                elif peeked_char == "\n":
                    # just let it keep reading
                    pass

            elif read_status == CurrentReadType.LookingForFieldMetaDataOrNextObject:

                if peeked_char == "\\":
                    token_builder = ""
                    read_status = CurrentReadType.ReadingFieldMetaData

                elif peeked_char == "\n":
                    # blank line will mean we are concluding this object
                    token_builder = ""
                    cur_object.fields.append(cur_field)
                    cur_group.objects.append(cur_object)
                    read_status = CurrentReadType.ReadAnything

            elif read_status == CurrentReadType.EncounteredComment_ReadToCR:

                # set the flag for reading the next line if necessary
                token_builder += just_read_char
                if peeked_char == "\n":
                    if revert_status_after_comment is not None:
                        read_status = revert_status_after_comment
                        revert_status_after_comment = None
                    else:
                        read_status = CurrentReadType.ReadAnything
                    if "IDD_Version" in token_builder:
                        self.idd.version_string = token_builder.strip().split(" ")[1].strip()
                        try:
                            version_tokens = self.idd.version_string.split(".")
                            tmp_string = "{}.{}".format(version_tokens[0], version_tokens[1])
                            self.idd.version_float = float(tmp_string)
                        except ValueError:
                            raise exceptions.ProcessingException(
                                "Found IDD version, but could not coerce into floating point representation")
                    elif "IDD_BUILD" in token_builder:
                        self.idd.build_string = token_builder.strip().split(" ")[1].strip()
                        magic_cache_key = "{}__{}".format(self.idd.version_string, self.idd.build_string)
                        module_logger.debug("Encountered IDD_BUILD, checking cache for key {}".format(magic_cache_key))
                        if magic_cache_key in IDD_CACHE:
                            module_logger.debug("Found this IDD cache key in the cache, using existing entry")
                            self.idd = IDD_CACHE[magic_cache_key]
                            return self.idd
                    token_builder = ""

        # end the file here, but should watch for end-of-file in other CASEs also
        self.idd.groups.append(cur_group)

        # we should assert that we have version and build strings, even in testing
        if (not self.idd.version_float) or (not self.idd.build_string):
            raise exceptions.ProcessingException("IDD did not appear to include standard version headers")

        # save this idd structure in the cache
        if magic_cache_key:
            IDD_CACHE[magic_cache_key] = self.idd
            module_logger.debug("Storing this IDD in cache with key: {}".format(magic_cache_key))

        # and return the magically useful IDDStructure instance
        return self.idd



from typing import List, Optional


class IDDField:
    """
    A simple class that defines a single field for an IDD object.  Relevant members are listed here:

    :ivar str field_an_index: Main identifier for this field
    :ivar dict(str,[str]) meta_data: A dictionary, where each key is a string metadata type, such as "\\note", and each
                                      value is a list of strings for each entry in the metadata of the key type.  So if
                                      the field has 3 note lines, the dictionary value for key "\\note" would be a 3
                                      element list, holding the 3 note lines.
    :ivar str field_name: A convenience variable holding the field name, if it is found in the metadata

    Constructor parameters:

    :param str an_index: The A_i or N_i descriptor for this field in the IDD, where i is an integer 1-...
    """

    def __init__(self, an_index: str):
        self.field_an_index = an_index
        self.meta_data = {}
        self.field_name: Optional[str] = None

    def __str__(self):
        return f"IDDField: {self.field_an_index} - {self.field_name}"


class IDDObject:
    """
    A simple class that defines a single IDD object.  Relevant members are listed here:

    :ivar str name: IDD Type, or name, of this object
    :ivar dict(str,[str]) meta_data: A dictionary, where each key is a string metadata type, such as "\\memo", and each
                                     value is a list of strings for each entry in the metadata of the key type.  So if
                                     the object has 3 memo lines, the dictionary value for key "\\memo" would be a 3
                                     element list, holding the 3 memo lines.
    :ivar list(IDDField) fields: A list of IDDField instances in order as read from the IDD

    Constructor parameters:

    :param str name: The object's type, or name
    """

    def __init__(self, name: str):
        self.name = name
        self.meta_data = {}
        self.fields: List[IDDField] = []

    def __str__(self):
        return f"IDDObject: {self.name} - {len(self.fields)} fields"


class IDDGroup:
    """
    A simple class that defines a single IDD group.  An IDD group is simply a container for IDD objects.
    Relevant members are listed here:

    :ivar str name: IDD Type, or name, of this group
    :ivar list(IDDObject) objects: A list of all objects found in the IDD within this group.

    Constructor parameters:

    :param str name: The group's name
    """

    def __init__(self, name: str):
        self.name = name
        self.objects: List[IDDObject] = []

    def __str__(self):
        return f"IDDGroup: {self.name} - {len(self.objects)} objects"


class IDDStructure:
    """
    An IDD structure representation.  This includes containing all the IDD objects (either inside groups or as
    standalone "single line objects"), as well as meta data such as the version ID for
    this IDD, and finally providing worker functions for accessing the IDD data

    Relevant "public" members are listed here:

    :ivar str file_path: The path given when instantiating this IDD, not necessarily an actual path
    :ivar float version_float: The floating point representation of the version of this IDD (for 8.6.0 it is 8.6)
    :ivar str build_string: The abbreviated git SHA used when generating this IDD
    :ivar [str] single_line_objects: A list of strings, each representing a raw, single-token, name-only IDD object
    :ivar list(IDDGroup) groups: A list of all groups found in the IDD, each of which will contain IDD objects

    Constructor parameters:

    :param str file_path: A file path for this IDD; not necessarily a valid path as it is never used, just available
                          for bookkeeping purposes.
    """

    def __init__(self, file_path: str):
        self.file_path = file_path
        self.version_string: Optional[str] = None
        self.build_string: Optional[str] = None
        self.version_float: Optional[float] = None
        self.single_line_objects: List[str] = []
        self.groups: List[IDDGroup] = []

    def get_object_by_type(self, type_to_get):
        """
        Given a type name, this returns the IDD object instance, or a single string if it is a single-line object

        :param type_to_get: The name of the object to get, case-insensitive as it is compared insensitively inside
        :return: If the object is a single-line object, simply the name; if the object is a full IDDObject instance,
                 that instance is returned.  If a match is not found, this returns None.
        """
        # check the normal objects
        for g in self.groups:
            for o in g.objects:
                if o.name.upper() == type_to_get.upper():
                    return o
        # check single line objects? Weird...but might be useful
        for o in self.single_line_objects:
            if o.upper() == type_to_get.upper():
                return o
        # if we haven't returned, fail
        return None

    def get_objects_with_meta_data(self, meta_data):
        """
        Given an object-level metadata string (\\required-object, e.g.), this returns objects that contain that metadata

        :param meta_data: An object-level metadata string, such as \\required-object
        :return: A list of IDDObjects that contain this metadata
        """
        objects = []
        for g in self.groups:
            for o in g.objects:
                if meta_data in o.meta_data:
                    objects.append(o)
        return objects
        # not going to look at single line objects for this



#!/usr/bin/env python

import codecs
import os
import setuptools

from pyiddidf import __version__


this_dir = os.path.abspath(os.path.dirname(__file__))
with codecs.open(os.path.join(this_dir, 'README.md'), encoding='utf-8') as i_file:
    long_description = i_file.read()

setuptools.setup(
    name='energyplus_idd_idf_utilities',
    version=__version__,
    packages=['pyiddidf'],
    description='EnergyPlus idd/idf manipulation in Python.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/myoldmopar/py-idd-idf',
    author='Edwin Lee',
    author_email='leeed2001@gmail.com',
    classifiers=[
        'Intended Audience :: Developers',
        'Topic :: Utilities',
        'License :: Public Domain',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],
    license='UnlicensedForNow',
    install_requires=[],
    entry_points={
        'console_scripts': ['energyplus_idd_idf=pyiddidf.cli:main_cli']
    }
)



import logging

module_logger = logging.getLogger("eptransition.idd.processor")


class ValidationIssue:
    """
    This class stores information about any issue that occurred when reading an IDF file.

    :param str object_name: The object type that was being validated when this issue arose
    :param int severity: The severity of this issue, from the class constants
    :param str message: A descriptive message for this issue
    :param str field_name: The field name that was being validated when this issue arose, if available.
    """

    INFORMATION = 0
    WARNING = 1
    ERROR = 2

    def __init__(self, object_name, severity, message, field_name=None):
        self.object_name = object_name
        self.severity = severity
        self.message = message
        self.field_name = field_name

    @staticmethod
    def severity_string(severity_integer):
        """
        Returns a string version of the severity of this issue

        :param int severity_integer: One of the constants defined in this class (INFORMATION, etc.)
        :return: A string representation of the severity
        """
        if severity_integer == ValidationIssue.INFORMATION:
            return "INFORMATION"
        elif severity_integer == ValidationIssue.WARNING:
            return "*WARNING*"
        elif severity_integer == ValidationIssue.ERROR:
            return "**ERROR**"
        else:
            raise Exception("Bad integer passed into severity_string()")

    def __str__(self):
        msg = " * Issue Found; severity = " + ValidationIssue.severity_string(self.severity) + "\n"
        msg += "  Object Name = " + self.object_name + "\n"
        if self.field_name is not None:
            msg += "  Field Name = " + self.field_name + "\n"
        return msg


class IDFObject(object):
    """
    This class defines a single IDF object.  An IDF object is either a comma/semicolon delimited list of actual
    object data, or a block of line delimited comments.  Blocks of comment lines are treated as IDF objects so they can
    be intelligently written back out to a new IDF file after transition in the same location.

    Relevant members are listed here:

    :ivar str object_name: IDD Type, or name, of this object
    :ivar [str] fields: A list of strings, one per field, found for this object in the IDF file

    Constructor parameters:

    :param [str] tokens: A list of tokens defining this idf object, the first token in the list is the object type.
    :param bool comment_blob: A signal that this list is comment data, and not an actual IDF object; default is False.
                              indicating it is meaningful IDF data.
    """

    def __init__(self, tokens, comment_blob=False):
        self.comment = comment_blob
        if comment_blob:
            self.object_name = "COMMENT"
            self.fields = tokens
        else:
            self.object_name = tokens[0]
            self.fields = tokens[1:]

    def __str__(self) -> str:
        return f"{self.object_name} : {len(self.fields)} fields"

    def object_string(self, idd_object=None):
        """
        This function creates an intelligently formed IDF object.  If the current instance is comment data, it simply
        writes the comment block out, line delimited, otherwise it writes out proper IDF syntax.  If the matching IDD
        object is passed in as an argument, the field names are matched from that to create a properly commented
        IDF object.

        :param IDDObject idd_object: The IDDObject structure that matches this IDFObject
        :return: A string representation of the IDF object or comment block
        """
        s = ""
        if self.comment:
            for comment_line in self.fields:
                s += comment_line + "\n"
            return s
        if not idd_object:
            if len(self.fields) == 0:
                s = self.object_name + ";\n"
            else:
                s = self.object_name + ",\n"
                padding_size = 25
                for index, idf_field in enumerate(self.fields):
                    if index == len(self.fields) - 1:
                        terminator = ";"
                    else:
                        terminator = ","
                    s += "  " + (idf_field + terminator).ljust(
                        padding_size) + "!- \n"
            return s
        else:
            if len(self.fields) == 0:
                s = self.object_name + ";\n"
            elif '\\format' in idd_object.meta_data and 'singleLine' in idd_object.meta_data['\\format']:
                field_token_string = ",".join([field for field in self.fields])
                s = self.object_name + ',' + field_token_string + ';\n'
            else:
                idd_fields = idd_object.fields
                s = self.object_name + ",\n"
                padding_size = 25
                for index, idf_idd_fields in enumerate(zip(self.fields, idd_fields)):
                    idf_field, idd_field = idf_idd_fields
                    if index == len(self.fields) - 1:
                        terminator = ";"
                    else:
                        terminator = ","
                    if "\\units" in idd_field.meta_data:
                        units_string = " {" + idd_field.meta_data["\\units"][0] + "}"
                    else:
                        units_string = ""
                    if idd_field.field_name is None:  # pragma no cover  - our files don't have an object like this yet
                        idd_field.field_name = ""
                    s += "  " + (str(idf_field) + terminator).ljust(
                        padding_size) + "!- " + idd_field.field_name + units_string + "\n"
            return s

    def validate(self, idd_object):
        """
        This function validates the current IDF object instance against standard IDD field tags such as minimum and
        maximum, etc.

        :param IDDObject idd_object: The IDDObject structure that matches this IDFObject
        :return: A list of ValidationIssue instances, each describing an issue encountered
        """
        issues = []
        # first thing check if we even have an IDD object to validate against
        if idd_object is None:
            module_logger.debug("Got \"None\" for idd_object when validating {}".format(self.object_name))
            return issues
        # then check some object level things
        if isinstance(idd_object, str):
            # we have a single-line string-only idd object, just leave
            return issues
        if "\\min-fields" in idd_object.meta_data:
            minimum_required_fields = int(idd_object.meta_data["\\min-fields"][0])
            actual_num_fields = len(self.fields)
            for i in range(minimum_required_fields):
                if i < actual_num_fields:  # if the item is there
                    if not self.fields[i]:  # if it's blank
                        if "\\default" in idd_object.fields[i].meta_data:  # if it's blank but has a default value
                            self.fields[i] = idd_object.fields[i].meta_data["\\default"][0]  # fill with default
                            # if it doesn't have a default, just leave it blank, later checks will catch it
                else:  # if the item isn't even there
                    if "\\default" in idd_object.fields[i].meta_data:  # if it has a default value
                        self.fields.append(idd_object.fields[i].meta_data["\\default"][0])  # fill with default
                    else:  # or if it doesn't have a default
                        self.fields.append("")  # make sure it does have an entry (blank) and it will be caught later
                        issues.append(ValidationIssue(idd_object.name, ValidationIssue.WARNING,
                                                      "Field within \\min-fields missing and no default",
                                                      idd_object.fields[i].field_name))
        for idf, idd in zip(self.fields, idd_object.fields):
            if "\\required-field" in idd.meta_data:
                if idf == "":
                    issues.append(ValidationIssue(idd_object.name, ValidationIssue.WARNING,
                                                  "Blank required field found", idd.field_name))
                    continue
            an_code = idd.field_an_index
            if an_code[0] == "N":
                if idf.strip() != "":
                    try:
                        number = float(idf)
                        if "\\maximum" in idd.meta_data:
                            max_constraint_string = idd.meta_data["\\maximum"][0]
                            if max_constraint_string[0] == "<":
                                max_val = float(max_constraint_string[1:])
                                if number >= max_val:
                                    issues.append(ValidationIssue(
                                        idd_object.name, ValidationIssue.WARNING,
                                        "Field value higher than idd-specified maximum>; actual={}, max={}".format(
                                            number, max_val), idd.field_name))
                            else:
                                max_val = float(max_constraint_string)
                                if number > max_val:
                                    issues.append(ValidationIssue(
                                        idd_object.name, ValidationIssue.WARNING,
                                        "Field value higher than idd-specified maximum; actual={}, max={}".format(
                                            number, max_val), idd.field_name))
                        if "\\minimum" in idd.meta_data:
                            min_constraint_string = idd.meta_data["\\minimum"][0]
                            if min_constraint_string[0] == ">":
                                min_val = float(min_constraint_string[1:])
                                if number <= min_val:
                                    issues.append(ValidationIssue(
                                        idd_object.name, ValidationIssue.WARNING,
                                        "Field value lower than idd-specified minimum<; actual={}, max={}".format(
                                            number, max_val), idd.field_name))
                            else:
                                min_val = float(min_constraint_string)
                                if number < min_val:
                                    issues.append(ValidationIssue(
                                        idd_object.name, ValidationIssue.WARNING,
                                        "Field value lower than idd-specified minimum; actual={}, max={}".format(
                                            number, max_val), idd.field_name))
                    except ValueError:
                        if "\\autosizable" in idd.meta_data and idf.upper() == "AUTOSIZE":
                            pass  # everything is ok
                        elif "\\autocalculatable" in idd.meta_data and idf.upper() in ["AUTOCALCULATE", "AUTOSIZE"]:
                            pass  # everything is ok
                        elif idf.upper() == "AUTOSIZE":
                            issues.append(ValidationIssue(
                                idd_object.name, ValidationIssue.WARNING,
                                "Autosize detected in numeric field that is _not_ listed autosizable", idd.field_name))
                        elif idf.upper() == "AUTOCALCULATE":
                            issues.append(ValidationIssue(
                                idd_object.name, ValidationIssue.WARNING,
                                "Autocalculate detected in numeric field that is _not_ listed autocalculatable",
                                idd.field_name))
                        else:
                            issues.append(ValidationIssue(
                                idd_object.name, ValidationIssue.WARNING,
                                "Non-numeric value in idd-specified numeric field", idd.field_name))
        return issues

    def write_object(self, file_object):
        """
        This function simply writes out the idf string to a file object

        :param file_object: A file-type object that responds to a write command
        :return: None
        """
        file_object.write(self.object_string())
        return None


class IDFStructure(object):
    """
    An IDF structure representation.  This includes containing all the IDF objects in the file, as well as meta data
    such as the version ID for this IDD, and finally providing worker functions for accessing the IDD data

    Relevant "public" members are listed here:

    :ivar str file_path: The path given when instantiating this IDF, not necessarily an actual path
    :ivar float version_float: The floating point representation of the version of this IDD (for 8.6.0 it is 8.6)
    :ivar [IDFObject] objects: A list of all IDF objects found in the IDF

    Constructor parameters:

    :param str file_path: A file path for this IDF; not necessarily a valid path as it is never used, just available
                          for bookkeeping purposes.
    """

    def __init__(self, file_path):
        self.file_path = file_path
        self.version_string = None
        self.version_float = None
        self.objects = None

    def get_idf_objects_by_type(self, type_to_get):
        """
        This function returns all objects of a given type found in this IDF structure instance

        :param str type_to_get: A case-insensitive object type to retrieve
        :return: A list of all objects of the given type
        """
        return [i for i in self.objects if i.object_name.upper() == type_to_get.upper()]

    def whole_idf_string(self, idd_structure=None):
        """
        This function returns a string representation of the entire IDF contents.  If the idd structure argument is
        passed in, it is passed along to object worker functions in order to generate an intelligent representation.

        :param IDDStructure idd_structure: An optional IDDStructure instance representing an entire IDD file
        :return: A string of the entire IDF contents, ready to write to a file
        """
        s = ""
        for idf_obj in self.objects:
            idd_obj = idd_structure.get_object_by_type(idf_obj.object_name)
            s += idf_obj.object_string(idd_obj) + "\n"
        return s

    def write_idf(self, idf_path, idd_structure=None):
        """
        This function writes the entire IDF contents to a file.  If the idd structure argument is
        passed in, it is passed along to object worker functions in order to generate an intelligent representation.

        :param str idf_path: The path to the file to write
        :param IDDStructure idd_structure: An optional IDDStructure instance representing an entire IDD file
        :return: None
        """
        with open(idf_path, "w") as f:
            f.write(self.whole_idf_string(idd_structure))
        return None

    def validate(self, idd_structure):
        """
        This function validates the current IDF structure instance against standard IDD object tags such as required
        and unique objects.

        :param idd_structure: An IDDStructure instance representing an entire IDD file
        :return: A list of ValidationIssue instances, each describing an issue encountered
        """
        issues = []
        required_objects = idd_structure.get_objects_with_meta_data("\\required-object")
        for r in required_objects:
            objects = self.get_idf_objects_by_type(r.name)
            if len(objects) == 0:
                issues.append(ValidationIssue(r.name, ValidationIssue.WARNING,
                                              message="Required object not found in IDF contents"))
        unique_objects = idd_structure.get_objects_with_meta_data("\\unique-object")
        for u in unique_objects:
            objects = self.get_idf_objects_by_type(u.name)
            if len(objects) > 1:
                issues.append(ValidationIssue(u.name, ValidationIssue.WARNING,
                                              message="Unique object has multiple instances in IDF contents"))
        for idf_object in self.objects:
            if idf_object.comment:
                continue
            idd_object = idd_structure.get_object_by_type(idf_object.object_name)
            this_object_issues = idf_object.validate(idd_object)
            if this_object_issues:
                issues.extend(this_object_issues)
        return issues

    def global_swap(self, dict_of_swaps):
        upper_case_swaps = {}
        for k, v in dict_of_swaps.items():
            upper_case_swaps[k.upper()] = v
        for idf_object in self.objects:
            if idf_object.comment:
                continue
            else:
                for i, idf_field in enumerate(idf_object.fields):
                    if idf_field.upper() in upper_case_swaps:
                        idf_object.fields[i] = upper_case_swaps[idf_field.upper()]
